from __future__ import annotations
from typing import Any, Union, Optional
from pathlib import Path
import inspect
from functools import wraps

import ray

from sinagot.workflow import Workflow
from sinagot.base import StepBase, SeedBase
from sinagot.logger import get_logger
from sinagot.storage import Storage

logger = get_logger()


class NodeMixin:
    name: str

    def get_storage(self, workflow: Workflow) -> Optional[Storage]:
        store_path = getattr(workflow.workspace, self.name, None)
        if store_path is not None:
            path = workflow._resolve_path(store_path)
            data_type = workflow.__annotations__[self.name]
            return Storage(
                path=path,
                workflow_id=workflow.workflow_id,
                data_type=data_type,
                read_kwargs=store_path.read_kwargs,
                write_kwargs=store_path.write_kwargs,
            )
        return None


class Step(StepBase, NodeMixin):
    def __init__(self, func: Any, *args: Any, **kwargs: Any):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.store_path: Union[None, Path] = None

    def __get__(self, workflow: Workflow, objtype: Any = None) -> Any:
        if workflow is None:
            return self
        return self.get_data(workflow, get_result=True)

    def decorate_func(self, workflow: Workflow) -> Any:
        storage = self.get_storage(workflow)

        @wraps(self.func)
        def decorated(*args: Any, **kwargs: Any) -> Any:
            logger.info(
                "Processing %s data for workflow %s", self, workflow.workflow_id
            )
            result = self.func(*args, **kwargs)

            if storage is not None:
                logger.info(
                    "Saving %s data for workflow %s", self, workflow.workflow_id
                )
                storage.write(result)

            return result

        return decorated

    def get_data(self, workflow: Workflow, get_result: bool = False) -> ray.ObjectRef:

        func = self.func
        kwargs = {}
        parameters = inspect.signature(func).parameters
        if "workflow_id" in parameters:
            kwargs["workflow_id"] = workflow.workflow_id

        storage = self.get_storage(workflow)
        if (storage is not None) and storage.exists():
            logger.debug(
                "Getting %s data for workflow %s from storage",
                self,
                workflow.workflow_id,
            )
            return storage.read()

        promise = workflow._run(self, func=self.decorate_func(workflow), **kwargs)
        if get_result:
            return ray.get(promise)
        return promise


def step(func: Any) -> Any:
    def step(*args: Any, **kwargs: Any) -> Step:
        return Step(func, *args, **kwargs)

    func.step = step
    return func


class Seed(SeedBase, NodeMixin):
    def __get__(self, workflow: Workflow, objtype: Any = None) -> Any:
        return self.get_data(workflow)

    def get_data(self, workflow: Workflow) -> ray.ObjectRef:
        storage = self.get_storage(workflow)
        if storage is not None:
            logger.debug("Getting %s data for workflow %s", self, workflow.workflow_id)
            return storage.read()


def seed() -> Any:
    return Seed()



import pkg_resources

from .workflow import Workflow
from .workspace import Workspace
from .node import step, seed
from .storage import LocalStorage

__all__ = [
    "local_file",
    "step",
    "Workspace",
    "Workflow",
    "seed",
    "LocalStorage",
]

__version__ = pkg_resources.get_distribution("sinagot").version



import logging

from sinagot.config import get_settings


def get_logger() -> logging.Logger:
    settings = get_settings()
    logger = logging.getLogger("sinagot")

    logger.setLevel(settings.LOGGING_LEVEL)
    if not logger.handlers:
        logger.addHandler(logging.StreamHandler())
    return logger



from typing import Type

import matplotlib.pyplot as plt
import networkx as nx

from sinagot.base import WorkflowBase


class WorkflowGraph:
    nx_graph: nx.Graph

    def __init__(self, workflow: Type[WorkflowBase]):
        self.workflow = workflow
        self._set_graph()

    def _set_graph(self) -> None:
        workflow = self.workflow
        graph = nx.DiGraph()
        graph.add_nodes_from(workflow._seeds)
        graph.add_nodes_from(workflow._steps)
        for name, step in workflow._steps.items():
            graph.add_edges_from(
                [
                    (source.name, name, {"label": step.func.__name__})
                    for source in step.args
                ]
            )
            graph.add_edges_from(
                [
                    (source.name, name, {"label": step.func.__name__})
                    for source in step.kwargs.values()
                ]
            )
        self.nx_graph = graph

    def draw(self) -> None:
        graph = self.nx_graph

        pos = nx.nx_pydot.pydot_layout(graph, prog="dot")
        edge_labels = {
            (source, dest): data["label"] for source, dest, data in graph.edges.data()
        }

        plt.figure(1, figsize=(20, 10))

        nx.draw_networkx(
            graph,
            pos=pos,
            arrows=True,
            node_color="#e1a798ff",
            font_color="#542418",
            font_weight="bold",
            node_shape="8",
            edge_color="#1a2127",
            node_size=2000,
            font_size=14,
            arrowsize=20,
        )
        nx.draw_networkx_edge_labels(
            graph,
            pos=pos,
            edge_labels=edge_labels,
            rotate=False,
            font_size=11,
            bbox={
                "boxstyle": "round",
                "ec": (1.0, 1.0, 1.0),
                "fc": (1.0, 1.0, 1.0),
                "alpha": 0.7,
            },
            alpha=0.7,
            # label_pos=0.6,
        )



from __future__ import annotations
from typing import Dict, Any, Tuple, Union
from pathlib import Path
import os

from sinagot.path_template import PathTemplate


class WorkflowBase:
    _seeds: Dict[str, Any]
    _steps: Dict[str, Any]


class AttrBase:
    def __set_name__(self, obj: Any, name: str) -> None:
        self.name = name

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.name})"


class SeedBase(AttrBase):
    """SeedBase"""


class StepBase(AttrBase):
    func: Any
    args: Tuple[Any, ...]
    kwargs: Dict[Any, Any]


class WorkspaceBase:

    WORKFLOW_ID_PATTERN = ".*"
    root_path: Path

    def _resolve_path(
        self, path: Union[str, Path, os.PathLike[str]], **kwargs: str
    ) -> PathTemplate:
        return PathTemplate(self.root_path, path).format(**kwargs)



from typing import Optional

from pydantic import BaseSettings


class Settings(BaseSettings):
    class Config:
        env_prefix = "SINAGOT_"

    LOGGING_LEVEL: str = "WARNING"
    RAY_LOCAL_MODE: bool = False
    RAY_LOGGING_LEVEL: str = "WARNING"
    RAY_ADDRESS: Optional[str]
    RAY_NUM_CPUS: Optional[int]
    RAY_NUM_GPUS: Optional[int]


def get_settings() -> Settings:
    return Settings()



from __future__ import annotations
from typing import Union, Any, Dict, Type, Optional, Generator
from pathlib import Path
import os
from glob import glob
import re

import pandas as pd

from sinagot.base import AttrBase, WorkspaceBase
from sinagot.logger import get_logger

logger = get_logger()


class LocalStorage(AttrBase):
    def __init__(
        self,
        path: Union[str, Path],
        read_kwargs: Optional[Dict[str, Any]] = None,
        write_kwargs: Optional[Dict[str, Any]] = None,
    ):
        self.path = path
        self.read_kwargs = read_kwargs or {}
        self.write_kwargs = write_kwargs or {}

    def __fspath__(self) -> str:
        return os.fspath(self.path)

    def iter_workflow_ids(self, workspace: WorkspaceBase) -> Generator[str, None, None]:
        workflow_ids = set()
        path_template = workspace._resolve_path(self.path)
        glob_pattern = path_template.format(workflow_id="*")
        re_pattern = str(
            path_template.format(workflow_id=f"({workspace.WORKFLOW_ID_PATTERN})")
        )
        for file_str in glob(str(glob_pattern)):
            match = re.match(re_pattern, file_str)
            if match:
                try:
                    workflow_id = match.group(1)
                    if workflow_id not in workflow_ids:
                        workflow_ids.add(workflow_id)
                        yield workflow_id
                except IndexError:
                    pass


class Storage:
    def __init__(
        self,
        workflow_id: str,
        path: Path,
        data_type: Any,
        read_kwargs: Dict[str, Any],
        write_kwargs: Dict[str, Any],
    ):
        self.workflow_id = workflow_id
        self.path = path
        self.data_type = data_type
        self.format = self.path.suffix
        self.read_kwargs = read_kwargs
        self.write_kwargs = write_kwargs

    def _get_handler(self) -> TypeHandler:
        try:
            return type_handlers[self.data_type](
                workflow_id=self.workflow_id,
                read_kwargs=self.read_kwargs,
                write_kwargs=self.write_kwargs,
            )
        except KeyError:
            raise NotImplementedError(f"{self.data_type.__name__} not implemented")

    def exists(self) -> bool:
        return self.path.exists()

    def read(
        self,
    ) -> Any:
        return self._get_handler().read(self.path)

    def write(self, data: Any) -> None:
        self._get_handler().write(self.path, data)


type_handlers: Dict[Any, Type[TypeHandler]] = {}


class TypeHandler:
    def __init_subclass__(cls, data_type: Any) -> None:
        type_handlers[data_type] = cls

    def __init__(
        self,
        workflow_id: str,
        read_kwargs: Dict[str, Any],
        write_kwargs: Dict[str, Any],
    ):
        self.workflow_id = workflow_id
        self.read_kwargs = read_kwargs
        self.write_kwargs = write_kwargs

    def read(self, path: Path) -> Any:
        """read"""

    def write(self, path: Path, data: Any) -> None:
        """write"""


class StringHandler(TypeHandler, data_type=str):
    def read(self, path: Path) -> Any:
        return path.read_text(**self.read_kwargs)

    def write(self, path: Path, data: Any) -> None:
        path.write_text(data, **self.write_kwargs)


class IntHandler(TypeHandler, data_type=int):
    def read(self, path: Path) -> Any:
        return int(path.read_text(**self.read_kwargs))

    def write(self, path: Path, data: Any) -> None:
        path.write_text(str(data), **self.write_kwargs)


class DataFrameHandler(TypeHandler, data_type=pd.DataFrame):
    def read(self, path: Path) -> Any:
        return pd.read_csv(path, **self.read_kwargs)

    def write(self, path: Path, data: Any) -> None:
        data.to_csv(path, **self.write_kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sinagot']

package_data = \
{'': ['*']}

install_requires = \
['networkx>=2.6.3,<3.0.0',
 'pandas>=1.3.5,<2.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'ray>=1.9.1,<2.0.0']

extras_require = \
{'draw': ['matplotlib>=3.5.1,<4.0.0', 'pydot>=1.4.2,<2.0.0']}

entry_points = \
{'console_scripts': ['test = tests.cli:run']}

setup_kwargs = {
    'name': 'sinagot',
    'version': '0.5.0',
    'description': 'Python lightweight workflow management framework with data exploration features',
    'long_description': '# Sinagot\n\n<p align="center">\n<a href="https://pypi.org/project/fastapi" target="_blank">\n    <img src="https://img.shields.io/pypi/v/sinagot?color=%2334D058&label=pypi%20package" alt="Package version">\n</a>\n<a href="https://pypi.org/project/fastapi" target="_blank">\n    <img src="https://img.shields.io/pypi/pyversions/sinagot.svg?color=%2334D058" alt="Supported Python versions">\n</a>\n</p>\n\n---\n\n**Source Code**: <a href="https://gitlab.com/YannBeauxis/sinagot" target="_blank">https://gitlab.com/YannBeauxis/sinagot</a>\n\n---\n\nSinagot is a Python lightweight workflow management framework using [Ray](https://www.ray.io/) as distributed computing engine.\n\nThe key features are:\n\n- **Easy to use**: Design workflow with simple Python classes and functions without external configuration files.\n- **Data exploration**: Access to computed data directly with object attributes, including complex type as pandas DataFrame.\n- **Scalable**: The [Ray](https://www.ray.io/) engine enable seamless scaling of workflows to external clusters.\n\n## Installation\n\n```bash\npip install sinagot\n```\n\n## Getting started\n\n```python\nimport pandas as pd\nimport sinagot as sg\n\n# Decorate functions to use them as workflow step\n@sg.step\ndef multiply(df: pd.DataFrame, factor: int) -> pd.DataFrame:\n    return df * factor\n\n\n@sg.step\ndef get_single_data(df: pd.DataFrame) -> int:\n    return int(df.iloc[0, 0])\n\n\n# Design a workflow\nclass TestWorkflow(sg.Workflow):\n    raw_data: pd.DataFrame = sg.seed() # seed is input data\n    factor: int = sg.seed()\n    multiplied_data: pd.DataFrame = multiply.step(raw_data, factor=factor)\n    final_data: int = get_single_data.step(multiplied_data)\n\n\n# Create a workspace on top of workflow for storage policy of data produced\nclass TestWorkspace(sg.Workspace[TestWorkflow]):\n    raw_data = sg.LocalStorage("raw_data/data-{workflow_id}.csv")\n    factor = sg.LocalStorage("params/factor")\n    multiplied_data = sg.LocalStorage(\n        "computed/multiplied_data-{workflow_id}.csv", write_kwargs={"index": False}\n    )\n    # In this example final_data is not stored and computed on demand\n\n\n# Create a workspace with local storage folder root path parameter\nws = TestWorkspace("/path/to/local_storage")\n\n# Access to a single workflow with its ID\nwf = ws["001"]\n\n# Access to item data, computed automatically if it does not exist in storage\ndisplay(wf.multiplied_data)\nprint(wf.final_data)\n```\n\nIn this example, the storage dataset is structured as follows :\n\n```\n├── params/\n│   └── factor\n├── raw_data/\n│   ├── data-{item_id}.csv\n│   └── ...\n└── computed/\n    ├── step-1-{item_id}.csv\n    └── ...\n```\n\nAnd the workflow is :\n\n<img src="docs/workflow.png" width="500">\n\n## Development Roadmap\n\nSinagot is at an early development stage but ready to be tested on actual datasets for workflows prototyping.\n\nFeatures development roadmap will be prioritized depending on usage feedbacks, so feel free to post an issue if you have any requirement.\n',
    'author': 'Yann Beauxis',
    'author_email': 'pro@yannbeauxis.net',
    'maintainer': 'Yann Beauxis',
    'maintainer_email': 'pro@yannbeauxis.net',
    'url': 'https://gitlab.com/YannBeauxis/sinagot',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations
import os
from pathlib import Path
from string import Formatter
from typing import Any, Iterator, Set, Dict, Iterable


class PathTemplate(Iterable[str]):
    def __init__(self, *args: Any, **kwargs: Any):
        self._path = Path(*args, **kwargs)

    def __repr__(self) -> str:
        return f"{type(self).__name__}('{os.fspath(self.path)}')"

    def __str__(self) -> str:
        return str(self.path)

    def __iter__(self) -> Iterator[str]:
        return iter(self.path.parts)

    @property
    def path(self) -> Path:
        return self._path

    def format(self, **kwargs: str) -> PathTemplate:
        path = Path(*(item.format_map(Default(**kwargs)) for item in self))
        return type(self)(path)

    def format_keys(self) -> Set[str]:
        return {
            i[1] for item in self for i in Formatter().parse(item) if i[1] is not None
        }


class Default(Dict[str, str]):
    def __missing__(self, key: str) -> str:
        return f"{{{key}}}"



from __future__ import annotations
from pathlib import Path
from typing import Iterator, Generator, Union, Type, Generic, TypeVar, Mapping, List

import ray

from sinagot.base import WorkflowBase, WorkspaceBase
from sinagot.logger import get_logger
from sinagot.config import get_settings
from sinagot.ray import init_ray

logger = get_logger()

IT = TypeVar("IT", bound=WorkflowBase)


class Workspace(Generic[IT], Mapping[str, IT], WorkspaceBase):

    Workflow: Type[IT]

    def __init__(self, *root_path_segments: Union[str, Path]):
        self.root_path = Path(*root_path_segments)
        self.settings = get_settings()
        init_ray()
        if type(self) != Workspace:
            logger.info("%s initialized", str(self))
            logger.debug("%s config :  %s", str(self), self.settings)

    def __repr__(self) -> str:
        return f"{type(self).__name__}('{self.root_path}')"

    @property
    def workflow_class(self) -> Type[IT]:
        return self.__orig_bases__[0]().__orig_class__.__args__[0]  # type: ignore

    def __getitem__(self, workflow_id: str) -> IT:
        return self.workflow_class(self, workflow_id)  # type: ignore

    def __iter__(self) -> Iterator[str]:
        yield from self._iter_workflow_ids()

    def _iter_workflow_ids(self) -> Generator[str, None, None]:
        ids = set()
        for name in self.workflow_class._seeds:
            seed = getattr(self, name)
            for workflow_id in seed.iter_workflow_ids(self):
                if workflow_id not in ids:
                    ids.add(workflow_id)
                    yield workflow_id

    def __len__(self) -> int:
        return len(list(self._iter_workflow_ids()))

    def remote(self, name: str) -> List[ray.ObjectRef]:
        return [
            getattr(type(workflow), name).get_data(workflow)
            for workflow in self.values()
        ]



from __future__ import annotations
from pathlib import Path
import os
import inspect
from typing import (
    Any,
    Union,
    Dict,
)

import ray

from sinagot.base import WorkflowBase, StepBase, SeedBase
from sinagot.path_template import PathTemplate
from sinagot.workflow_graph import WorkflowGraph
from sinagot.workspace import Workspace
from sinagot.logger import get_logger


logger = get_logger()


class Workflow(WorkflowBase):
    _seeds: Dict[str, Any]
    _steps: Dict[str, Any]
    workflow_id: str
    graph: WorkflowGraph

    def __init__(self, workspace: Workspace[Workflow], workflow_id: str):
        self.workspace = workspace
        self.workflow_id = workflow_id

    def __init_subclass__(cls, **kwargs: Any):
        cls._seeds = dict(inspect.getmembers(cls, predicate=is_seed))
        cls._steps = dict(inspect.getmembers(cls, predicate=is_step))
        cls.graph = WorkflowGraph(cls)

    def __repr__(self) -> str:
        return f"{type(self).__name__}('{self.workflow_id}')"

    def _resolve_path(
        self, path: Union[str, Path, os.PathLike[Any]], **kwargs: str
    ) -> Path:
        kwargs["workflow_id"] = self.workflow_id
        path_ = os.fspath(path)
        return PathTemplate(self.workspace.root_path, path_).format(**kwargs).path

    def _run(self, step: StepBase, func: Any, **params: Any) -> ray.ObjectRef:
        logger.debug("Running %s for workflow %s", step, self.workflow_id)
        args = [self._get_node(source.name) for source in step.args]
        kwargs = {
            name: self._get_node(source.name) for name, source in step.kwargs.items()
        }
        remote_func = ray.remote(func)
        return remote_func.remote(*args, **kwargs, **params)

    def _get_node(self, name: str) -> Any:
        return type(self).__dict__[name].get_data(self)


def is_step(attribute: Any) -> bool:
    return isinstance(attribute, StepBase)


def is_seed(attribute: Any) -> bool:
    return isinstance(attribute, SeedBase)



import ray

from sinagot.config import get_settings
from sinagot.logger import get_logger

RAY_PREFIX = "RAY_"


def init_ray() -> None:

    if not ray.is_initialized():
        settings = get_settings()
        prefix_len = len(RAY_PREFIX)
        kwargs = {
            key.lower()[prefix_len:]: value
            for key, value in settings.dict().items()
            if key.upper().startswith(RAY_PREFIX) and (value is not None)
        }

        init_message = ray.init(ignore_reinit_error=True, **kwargs)
        logger = get_logger()
        logger.debug("ray init %s", init_message)
